#!/usr/bin/perl -Tw
#####################################################################
#
# WakeMeUp - an alternative (VDR) - wakeup solution with WoL
# (server)
#
# (c)2008 Matthias Eckert <matzetronic at gmx dot de>
#
# Version 0.13  (2008-02-23)
#
#####################################################################


use strict;
use warnings;

# predefined error code constants
use constant { ERROR => 0, TRUE => 1, FALSE => 2, PARPROB => 3, UNSET => 4 };

use IO::Socket;
use Sys::Syslog;
use Getopt::Long;
use Fcntl qw(:flock);

my $PROGNAME = "WakeMeUp server";
my $PROGVERS = "0.13";
my $PROGDATE = "2008-02-23";


my %Config = ();
my %Jobs = ();
my $status;
my $msg;
my $pidFlag = 0;
my $remoteHost;
my $macAddress;
my $time1;
my $time2;
my $sisPM;
my $sisDevice;
my $sisPort;
my $diff;
my $method;
my $getup;
my $message;
my $psname;



sub usage {
    print STDOUT << "EOL";

Usage: wakemeup-server [options]

Available options:
  --help                         Print this help message.
  --bind-address=<ip address>    IP address to bind service to (default is 0.0.0.0).
  --bind-port=<port number>      Port number to bind service to (default is 9999).
  --syslog                       Log all output via syslogd instead of stdout/stderr.
  --wakeupbefore=<minutes>       Start that many minutes before the wakeuptime (default is 5).
  --daemonize                    Causes the server to operate in the background.
  --nicelevel=<level>            Run with an adjusted niceness (default is 15).


EOL
;
    exit 0;
}



sub parse_cmdline {
    my $result;
    
    &Getopt::Long::Configure('pass_through', 'prefix_pattern=--');
    $result = GetOptions (  'help'		=> \&usage,
			    'bind-address=s'	=> \$Config{IP},
			    'bind-port=i'	=> \$Config{Port},
			    'syslog'		=> \$Config{Syslog},
			    'wakeupbefore=i'	=> \$Config{Before},
			    'daemonize'		=> \$Config{Daemon},
			    'nicelevel=i'	=> \$Config{Priority}
			 );

    if ($#ARGV > -1) {
        # unknown options
        foreach (@ARGV) {
            print STDOUT "Unknown command line option '$_'.\n";
        }
        print STDOUT "See 'wakemeup-server --help' for a list of available options.\n";
        exit 1;
    }
    # check bind address
    if (defined ($Config{IP})) {
        if (! $Config{IP} || $Config{IP} !~ /^(([01]?[0-9][0-9]?|2[0-4][0-9]|25[0-5])\.){3}([01]?[0-9][0-9]?|2[0-4][0-9]|25[0-5])$/) {
	    print STDERR "Invalid value for '--bind-address'!\n";
	    exit 1;
	}
    }
    else {
        # listen on all interfaces as default
        $Config{IP} = "0.0.0.0";
    }
    # check bind port
    if (defined ($Config{Port})) {
        if (! $Config{Port} || $Config{Port} !~ /^[\d]{1,5}$/) {
	    print STDERR "Invalid value for '--bind-port'!\n";
	    exit 1;
	}
	if ($Config{Port} < 1 || $Config{Port} > 65535) {
	    print STDERR "Invalid value for '--bind-port'!\n";
	    exit 1;
	}
    }
    else {
        # port 9999 as default
        $Config{Port} = "9999";
    }
    # check nice level
    if (defined ($Config{Priority})) {
        if (! $Config{Priority} || $Config{Priority} !~ /^(\-)?[\d]{1,2}$/) {
	    print STDERR "Invalid value for '--nicelevel'!\n";
	    exit 1;
	}
	if ($Config{Priority} < -20 || $Config{Priority} > 19) {
	    print STDERR "Invalid value for '--nicelevel'!\n";
	    exit 1;
	}
    }
    else {
        # low priority (15) as default
        $Config{Priority} = 15;
    }
    # check wakeup before
    if (defined ($Config{Before})) {
        if (! $Config{Before} || $Config{Before} !~ /^[\d]{1,3}$/) {
	    print STDERR "Invalid value for '--wakeupbefore'!\n";
	    exit 1;
	}
	if ($Config{Before} < 1 || $Config{Before} > 60) {
	    print STDERR "Invalid value for '--wakeupbefore'!\n";
	    exit 1;
	}
    }
    else {
        # 5 minutes as default
        $Config{Before} = "5";
    }
    return TRUE;
}



sub find_executables {
    my $path = "/usr/local/bin:/usr/bin:/bin:/usr/sbin:/sbin";
    my $dir;
    my $cmd;
    
    foreach $cmd (qw/arp at wakeonlan wakeup sispmctl/) {
        foreach $dir (split(/:/, $path)) {
            if (-x "$dir/$cmd" && ! defined $Config{$cmd}) {
                $Config{$cmd} = "$dir/$cmd";
                $Config{$cmd} =~ /(.*)/;        # evil untaint
                $Config{$cmd} = $1;
                last;
            }
        }
    }
}



sub logger {
    my $msg = shift || return ERROR;
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$ydat,$isdst) = localtime();
    my $date = sprintf ("%4d-%02d-%02d %02d:%02d:%02d", $year+1900,$mon+1,$mday,$hour,$min,$sec);
    
    $msg =~ s/[\r\n]+//g;
    if (defined ($msg) && $msg) {
        if (defined ($Config{Syslog})) {
            openlog "WakeMeUp ", "pid", "user";
            syslog ('user|info', "$msg");
            closelog;
        }
        else {
	    if (! defined ($Config{Daemon})) {
                print STDOUT "[$date]  $msg\n";
	    }
        }
    }    
    return TRUE;
}



sub pidfile {
    my $action = shift;
    my $pidfile = "/var/run/wakemeup-server.pid";
    my $pid;
    
    if (defined ($action) && $action) {
        if ($action eq "create") {
            # pid file exists already
            if (-f $pidfile) {
                # open the pid file for reading
                if (open (PID, "$pidfile")) {
                    flock (PID, LOCK_SH);
                    chomp ($pid = <PID>);
                    flock (PID, LOCK_UN);
                    close PID;
                    # empty pid file -> abort
                    if (! defined ($pid) || ! $pid) {
                        return (ERROR, "Pid file '$pidfile' exists already!");
                    }
                    else {
                        # untaint for $pid
                        $pid =~ /^([\d]{1,5})$/;
                        $pid = $1;
                        # check variable again
                        if (! defined ($pid) || ! $pid) {
                            return (ERROR, "Pid file '$pidfile' exists already!");
                        }
                        # is there a running process with this pid ?
                        if (kill (0, $pid)) {
                            return (ERROR, "Another instance is already running with pid $pid!");
                        }
                        else {
                            # stale pid file, open the pid file for read/write
                            if (open (PID, "> $pidfile")) {
                                flock (PID, LOCK_EX);
                                print PID "$$\n";
                                flock (PID, LOCK_UN);
                                close PID;
                                $pidFlag = 1;
                                return (TRUE, "Pid file '$pidfile' created!");
                            }
                            else {
                                # could not overwrite pid file
                                return (ERROR, "Pid file '$pidfile' exists already!");
                            }
                        }
                    }
                }
                else {
                    # could not open pid file for reading
                    return (ERROR, "Pid file '$pidfile' exists already!");
                }
            }
            else {
                # pid file does not exist, create it
                if (open (PID, "> $pidfile")) {
                    flock (PID, LOCK_EX);
                    print PID "$$\n";
                    flock (PID, LOCK_UN);
                    close PID;
                    $pidFlag = 1;
                    return (TRUE, "Pid file '$pidfile' created!");
                }
                else {
                    # could not write pid file
                    return (ERROR, "Couldn't open pid file '$pidfile' for writing!");
                }
            }
        }
        elsif ($action eq "delete") {
            if ($pidFlag) {
                if (-f $pidfile) {
                    unlink $pidfile;
                    return (TRUE, "Pid file '$pidfile' deleted!");
                }
                else {
                    return (FALSE, "Pid file '$pidfile' already deleted!?");
                }
            }
            else {
                return (FALSE, "Pid file not created, therefore not deleted!");
            }
        }
        else {
            return (PARPROB, "Invalid number of arguments!");
        }
    }
    else {
        return (PARPROB, "No arguments given!");
    }
}



sub exit_ {
    my $exit_code = shift;

    &pidfile("delete");
    if (defined ($exit_code) && $exit_code && $exit_code =~ /^[\d]{1,3}/ && $exit_code < 255) {
        exit $exit_code;
    }
    else {
        &logger("Server v$PROGVERS stopped.");
        exit 0;
    }
}



sub agent_smith {
    my $pid;
    my ($status, $msg);
    
    if (defined ($pid = fork())) {
        if ($pid) {
            # parent, to be exited
            exit 0;
        }
        else {
            # child, keep running
            ($status, $msg) = &pidfile("create");
            if ($status != TRUE) {
                print STDERR "$msg\n";
                &exit_(1);
            }
        }
    }
    else {
        print STDERR "Unable to fork!\n";
        &exit_(1);
    }
}



sub read_jobs {
    my ($wakeup, $time, $ip, $mac);
    my $jobid;
    my $magic_RE = qr/^\#\#\# WakeMeUp at job \|/;
    my @list = ();
    my @content = ();
    my $line;
    my $dummy;
    my $max;
    my $i;
    my $status;
    
    # get job listing
    chomp (@list = `$Config{at} -l 2>/dev/null`);
    $status = $?;
    if (defined ($status)) {
        # exit status from at command is 0 - success
        if ($status == 0) {
            %Jobs = ();
            $max = @list;
            # are there any jobs ?
            if ($max) {
                for ($i = 0; $i < $max; $i++) {
                    # check for valid job id
                    if ($list[$i] =~ /^([\d]+)[\s\t]+(Mon|Tue|Wed|Thu|Fri|Sat|Sun)[\s\t]+.*/) {
                        $jobid = $1;
                        if ($jobid) {
                            # get details for job with this id
                            chomp (@content = `$Config{at} -c $jobid 2>/dev/null`);
                            $status = $?;
                            if (defined ($status)) {
                                # exit status from at command is 0 - success
                                if ($status == 0) {
                                    # find the magic line :-)
                                    foreach $line (@content) {
                                        next if ($line !~ /$magic_RE/);
                                        if ($line =~ /$magic_RE/) {
                                            # split into values
                                            ($dummy, $wakeup, $time, $ip, $mac) = split (/\|/, $line, 5);
                                            if (defined ($wakeup) && defined ($time) && defined ($ip) && defined ($mac)) {
                                                # write values into job hash
                                                $Jobs{$jobid}{Plain} = "$wakeup|$time|$ip|$mac";
                                                $Jobs{$jobid}{Wakeup} = $wakeup;
                                                $Jobs{$jobid}{Time} = $time;
                                                $Jobs{$jobid}{IP} = $ip;
                                                $Jobs{$jobid}{MAC} = $mac;
                                            }
                                            # magic line found, so we can leave this loop
                                            last;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    # Note: no error handling in this function - this is ok here !
    return TRUE;
}



sub add_job {
    my ($wakeup, $time, $ip, $mac, $minutes, $device, $port) = @_;
    my $jobid;
    my $magic = '### WakeMeUp at job |';
    my @old = ();
    my $is_added = 0;
    my $command;
    
    # clear path environment as described in the "Perl Taintmode FAQ"
    $ENV{'PATH'} = "";
    # set path environment
    $ENV{'PATH'} = "/usr/local/bin:/usr/bin:/bin:/usr/sbin:/sbin";
    # untaint for variables
    $ip =~ /^([\d]{1,3}\.[\d]{1,3}\.[\d]{1,3}\.[\d]{1,3})$/;
    $ip = $1;
    $mac =~ /^(([0-9A-F]{2,2}:){5,5}[0-9A-F]{2,2})$/i;
    $mac = $1;
    $minutes =~ /^([\d]+)$/;
    $minutes = $1;
    $wakeup =~ /^([\d]+)$/;
    $wakeup = $1;
    $time =~ /^([\d]+)$/;
    $time = $1;
    if (defined ($device) && defined ($port)) {
        $device =~ /^(\d)$/;
        $device = $1;
        $port =~ /^(\d)$/;
        $port = $1;
    }
    # check variables
    if (! defined ($ip) || ! defined ($mac) || ! defined ($minutes) || ! defined ($wakeup) || ! defined ($time)) {
        return (ERROR, "Something wrong!");
    }

    # read all jobs into job hash
    &read_jobs;
    # search for existing entries from this host
    foreach $jobid (keys %Jobs) {
        # search for existing entry with the same content
        if ($Jobs{$jobid}{Plain} eq "$wakeup|$time|$ip|$mac") {
            # already added, set flag to TRUE
            $is_added = 1;
        }
        # search for other existing entries from this host
        elsif ($Jobs{$jobid}{IP} eq "$ip" && $Jobs{$jobid}{MAC} eq "$mac" && $Jobs{$jobid}{Plain} ne "$wakeup|$time|$ip|$mac") {
            # old job - mark for delete
            push (@old, $jobid);
        }
    }
    # check values in $minutes, $wakeup and $time - delete all timers for this host, if set to zero
    if (! $minutes && ! $wakeup && ! $time) {
        foreach (@old) {
            if (&delete_job($_)) {
                &logger("Removing old job with id $_ for host $ip...");
            }
        }
        # return
        return (UNSET, 0);
    }
    # job already added, so we can delete the other old jobs if any
    if ($is_added) {
        foreach (@old) {
            if (&delete_job($_)) {
                &logger("Removing old job with id $_ for host $ip...");
            }
        }
        # return
        return (FALSE, $wakeup);
    }
    # job not yet added, do it now
    else {
        # Sis-PM
        if (defined ($device) && defined ($port) && defined ($Config{sispmctl})) {
            if ($minutes < 2) {
                $minutes += 1;
                $wakeup += 60;
            }
            $command = "/bin/echo -e \"$magic$wakeup|$time|$ip|$mac\n$Config{sispmctl} -d $device -o $port\" | $Config{at} now + $minutes minutes 2>/dev/null";
            $minutes -= 1;
            $wakeup -= 60;
            $command .= " && /bin/echo -e \"$magic$wakeup|$time|$ip|$mac\n$Config{sispmctl} -d $device -f $port\" | $Config{at} now + $minutes minutes 2>/dev/null";
        }
        # WoL
        elsif (defined ($Config{arp}) && defined ($Config{wol})) {
            # log, if WoL is our fallback
            if (defined ($device) && defined ($port) && ! defined ($Config{sispmctl})) {
                &logger("Warning: Command 'sispmctl' not found, using Wake-on-Lan instead!");
            }
            $command = "/bin/echo -e \"$magic$wakeup|$time|$ip|$mac\n$Config{arp} -n -s $ip $mac ; $Config{wol} -i $ip $mac ; $Config{arp} -n -d $ip\" | $Config{at} now + $minutes minutes 2>/dev/null";
        }
        else {
            return (ERROR, "Couldn't add WakeUp time!");
        }
        if (!system $command) {
            # only delete old jobs if the new one can be set
            foreach (@old) {
                if (&delete_job($_)) {
                    &logger("Removing old job with id $_ for host $ip...");
                }
            }
            # all done, return
            return (TRUE, $wakeup);
        }
        else {
            # hmm, something wrong - return without deleting old jobs
            return (ERROR, "Couldn't add WakeUp time!");
        }
    }
}



sub delete_job {
    my $jobid = shift;
    
    if (! system "$Config{at} -d $jobid 2>/dev/null") {
        return (TRUE, "Job with id $jobid deleted!");
    }
    else {
        return (ERROR, "Couldn't delete job with id $jobid!");
    }
}



sub no_wakeup {
    my $ip = shift;
    my $mac = shift;
    
    if (defined ($ip) && defined ($mac) && $ip && $mac) {
        ($status, $msg) = &add_job(0, 0, $ip, $mac, 0);
        return ($status, $msg);
    }
    else {
        return (ERROR, "Something wrong!");
    }
}



sub wakeup1 {
    my $ip = shift;
    my $mac = shift;
    my $time = shift;
    my $device = shift;
    my $port = shift;
    my $before;
    my $now;
    my $wakeup;
    my $minutes;
    my ($status, $msg);
    
    if (defined ($ip) && defined ($mac) && defined ($time) && $ip && $mac && $time && $time =~ /^[\d]{10,10}$/) {
        $before = int($Config{Before} * 60);
        # calculate timestamp for wakeup
        $wakeup = int($time - $before);
        # set the seconds to 00
        $wakeup = int($wakeup - ($wakeup % 60));
        # get the actual time
        $now = int(time() + 1);
        # is the wakeup timestamp in future ?
        if ($now < $wakeup) {
            # calculate difference from now in minutes
            $minutes = int((($wakeup - $now) / 60) + 1);
            # set the seconds for $time to 00
            $time = int($time - ($time % 60));
	    # add the wakeup job
            ($status, $msg) = &add_job($wakeup, $time, $ip, $mac, $minutes, $device, $port);
            return ($status, $msg);
        }
        elsif ($now <= $time) {
            return (ERROR, "Timestamp value is too small!");
        }
	else {
	    return (ERROR, "Timestamp value is in the past!?");
	}
    }
    else {
        return (ERROR, "Something wrong!");
    }
}



sub wakeup2 {
    my $ip = shift;
    my $mac = shift;
    my $time = shift;
    my $rnow = shift;
    my $device = shift;
    my $port = shift;
    my $before;
    my $diff;
    my $delta;
    my $now;
    my $wakeup;
    my $minutes;
    
    if (defined ($ip) && defined ($mac) && defined ($time) && defined ($rnow) && $ip && $mac && $time && $rnow && $time =~ /^[\d]{10,10}$/ && $rnow =~ /^[\d]{10,10}$/) {
        $before = int($Config{Before} * 60);
        # check timestamps
        if (int($time - $before) > $rnow) {
            # calculate the difference between client time and client wakeup
            $diff = int($time - $rnow);
            # get the actual time
            $now = int(time() + 1);
            # calculate timestamp for wakeup
            $wakeup = int(($now + $diff) - $before);
            # set the seconds to 00
            $wakeup = int($wakeup - ($wakeup % 60));
            # calculate the delta
            if ($rnow > $now) {
                $delta = int($rnow - $now);
            }
            elsif ($now > $rnow) {
                $delta = int($now - $rnow);
            }
            else {
                $delta = 0;
            }
            if ($now < $wakeup) {
                # calculate difference from now in minutes
                $minutes = int((($wakeup - $now) / 60) + 1);
                # set the seconds for $time to 00
                $time = int($time - ($time % 60));
	        # add the wakeup job
                ($status, $msg) = &add_job($wakeup, $time, $ip, $mac, $minutes, $device, $port);
                return ($status, $msg, $delta);
            }
            elsif ($time >= $rnow) {
                return (ERROR, "Timestamp value is too small!", 0);
            }
            else {
                return (ERROR, "Something wrong!", 0);
            }
        }
        elsif ($time >= $rnow) {
            return (ERROR, "Timestamp value is too small!", 0);
        }
        else {
            return (ERROR, "Timestamp value is in the past!?", 0);
        }
    }    
    else {
        return (ERROR, "Something wrong!", 0);
    }
}



sub get_mac {
    my $ip = shift;
    my ($ipaddr, $hwtype, $flags, $hwaddr, $mask, $device);
    my @arp;
    
    if (defined ($ip) && $ip && $ip =~ /^[\d]{1,3}\.[\d]{1,3}\.[\d]{1,3}\.[\d]{1,3}$/) {
        if (open(FH, "/proc/net/arp")) {
	    chomp (@arp = <FH>);
	    close FH;
	    if (@arp) {
	        foreach (@arp) {
		    ($ipaddr, $hwtype, $flags, $hwaddr, $mask, $device) = split (/[\s\t]+/, $_, 6);
		    if (defined ($ipaddr) && defined ($hwaddr)) {
		        if ($ipaddr eq $ip && $hwaddr =~ /^([0-9A-F]{2,2}:){5,5}[0-9A-F]{2,2}$/i) {
			    return (TRUE, $hwaddr);
			}
		    }
		}
		return (FALSE, "Couldn't find any matching arp entry!");
	    }
	    else {
	        return (ERROR, "No arp entries!");
	    }
	}
	else {
	    return (ERROR, "Couldn't read arp table!");
	}
    }
    else {
        return (ERROR, "Something wrong!");
    }
}



####################
#####   MAIN   #####
####################

# change the process name
$psname = $0;
foreach (@ARGV) {
    $psname .= " $_";
}
$0 = $psname;

print STDOUT "$PROGNAME $PROGVERS ($PROGDATE) by Matthias Eckert\n\n";

# parse commandline arguments
&parse_cmdline;

# check for uid 0
if ( $> != 0 ) {
    print STDERR "Sorry, this program must be started as root!\n";
    print STDERR "(This may be changed in future versions)\n";
    exit 1;
}

# check for executables
&find_executables;
# check 'at'
if (! defined ($Config{at})) {
    print STDERR "Can't find 'at' command!\n";
    exit 1;
}
# check 'wol'
if (defined ($Config{wakeonlan})) {
    $Config{wol} = $Config{wakeonlan};
}
elsif (defined ($Config{wakeup})) {
    $Config{wol} = $Config{wakeup};
}
# check (arp && wol) OR (sispmctl)
if (! defined ($Config{sispmctl}) && (! defined $Config{arp}) || ! defined ($Config{wol})) {
    if (! defined ($Config{arp})) {
        print STDERR "Can't find 'arp' command!\n";
    }
    if (! defined ($Config{wol})) {
        print STDERR "Can't find 'wakeonlan' / 'wakeup' command!\n";
    }
    if (! defined ($Config{sispmctl})) {
        print STDERR "Can't find 'sispmctl' command!\n";
    }
    exit 1;
}


# set signal handlers to 'ignore' during startup
local $SIG{'INT'}  = 'IGNORE';
local $SIG{'TERM'} = 'IGNORE';

# fork()
if (defined ($Config{Daemon})) {
    &agent_smith;
}
else {
# not running in background
    ($status, $msg) = &pidfile("create");
    if ($status != TRUE) {
        print STDERR "$msg\n";
        &exit_(1);
    }
}

# set signal handlers - needed for cleanup
local $SIG{'INT'}  = sub {&exit_(0);};
local $SIG{'TERM'} = sub {&exit_(0);};

# set up the server socket
my $server = IO::Socket::INET->new ( Proto => "tcp",
                                     LocalHost => $Config{IP},
                                     LocalPort => $Config{Port},
                                     Listen    => SOMAXCONN,
                                     Reuse     => 1);
if (! defined ($server) || ! $server) {
    print STDERR "Can't setup socket for WakeMeUp server: $!\n";
    &exit_(1);
}

# print security warning
if ("$Config{IP}" eq "0.0.0.0") {
    print STDERR " Warning: Server will listen on all available interfaces! This could be a security risk!\n\n";
}

# print warning about nice level
if ($Config{Priority} < 0) {
    print STDERR " Note: Running the server with an high process priority is not recommended!\n\n";
}

&logger("Server v$PROGVERS started.");

# set the process priority
setpriority 0, 0, $Config{Priority};

# now enter the loop
while (my $client = $server->accept()) {
    $client->autoflush(1);
    $remoteHost = undef;
    $remoteHost = $client->peerhost;
    # w00t ! really connected ?
    next if (! defined($remoteHost));
    &logger("[$remoteHost]  Connect.");
    my $line = <$client>;
    if (defined ($line)) {
        # remove non-printable chars
        $line =~ s/[^\x20-\x7e]//g;
        # replace some whitespaces with only one
        $line =~ s/[\s\t]{2,}/\ /g;
        if ($line =~ /^([\d]{10,10}|0)(\s\d:\d)?$/ || $line =~ /^([\d]{10,10}|0)\s[\d]{10,10}(\s\d:\d)?$/) {
	    ($status, $msg) = &get_mac($remoteHost);
	    if ($status == TRUE) {
                $macAddress = undef;
	        $macAddress = $msg;
		if ($macAddress) {
                    ($time1, $time2) = undef;
                    ($sisPM, $sisDevice, $sisPort) = undef;
                    # using method 1
                    # WoL
                    if ($line =~ /^[\d]{10,10}$/) {
                        $time1 = $line;
                        $method = 1;
                        ($status, $msg) = &wakeup1($remoteHost, $macAddress, $time1);
                    }
                    # Sis-PM
                    elsif ($line =~ /^[\d]{10,10}\s\d:\d$/) {
                        ($time1, $sisPM) = split (/\s/, $line);
                        ($sisDevice, $sisPort) = split (/:/, $sisPM);
                        $method = 1;
                        ($status, $msg) = &wakeup1($remoteHost, $macAddress, $time1, $sisDevice, $sisPort);
                    }
                    # using method 2
                    # WoL
                    elsif ($line =~ /^[\d]{10,10}\s[\d]{10,10}$/) {
                        ($time1, $time2) = split (/\s/, $line);
                        $method = 2;
                        ($status, $msg, $diff) = &wakeup2($remoteHost, $macAddress, $time1, $time2);
                    }
                    # Sis-PM
                    elsif ($line =~ /^[\d]{10,10}\s[\d]{10,10}\s\d:\d$/) {
                        ($time1, $time2, $sisPM) = split (/\s/, $line);
                        ($sisDevice, $sisPort) = split (/:/, $sisPM);
                        $method = 2;
                        ($status, $msg, $diff) = &wakeup2($remoteHost, $macAddress, $time1, $time2, $sisDevice, $sisPort);
                    }
                    # no new timer - delete all timer for this host
                    elsif ($line =~ /^0(\s\d:\d)?$/ || $line =~ /^0\s[\d]{10,10}(\s\d:\d)?$/) {
                        ($status, $msg) = &no_wakeup($remoteHost, $macAddress);
                    }
                    # hmm, impossible error...
                    else {
                        $status = ERROR;
                        $msg = "Couldn't get timestamp!";
                    }
                    # status: added
                    if ($status == TRUE) {
                        $getup = scalar localtime($msg);
                        $message = "OK. Get up at: $getup";
                        if (defined ($sisDevice) && defined ($sisPort)) {
                            $message .= " (Sis-PM)";
                        }
                        else {
                            $message .= " (WoL)";
                        }
                        if ($method == 2 && $diff > 86400) {
                            $message .= "   +++WARNING: Time drift is >24 hours!+++";
                        }
                    }
                    # status: already added
                    elsif ($status == FALSE) {
                        $getup = scalar localtime($msg);
                        $message = "OK. Get up at: $getup (already added)";
                        if ($method == 2 && $diff > 86400) {
                            $message .= "   +++WARNING: Time drift is >24 hours!+++";
                        }
                    }
                    # status: nothing set
                    elsif ($status == UNSET) {
                        $message = "OK. Nothing set.";
                    }
                    # status: not added
                    else {
                        $message = $msg;
                    }
		}
		else {
                    $message = "Couldn't get mac address!";
		}
	    }
	    else {
                $message = $msg;
	    }
        }
        else {
            $message = "Wrong input!";
        }
    }
    else {
        $message = "Wrong input!";
    }
    &logger("[$remoteHost]  $message");
    print $client "$message\r\n";
    &logger("[$remoteHost]  Disconnect.");
    # closing connection
    $client->close();
}

&exit_(0);




#####################################################################
#
# History:
#
# Version 0.13   (2008-02-23) me
# - added (experimental) support for USB Outlet SIS PM / Intelli-Plug
#
# Version 0.12   (2008-02-17) me
# - fixed a typo in commandline option '--nicelevel'
# - fixed a typo in commandline option '--wakeupbefore'
# - changed logging of pid in function 'logger', because of warnings
#   on some systems
#
# Version 0.11   (2008-02-12) me
# - fixed a typo
#
# Version 0.10   (2008-02-12) me
# - added '/sbin' to path environment (thx to [magicamun])
# - added function 'find_executables' to find our needed executables,
#   this should fix some path problems (e.g. on SuSE)
#   (reported by [magicamun])
# - added support for 'wakeup' script (for systems without 'wakeonlan'
#   program :-)
#
# Version 0.9   (2008-02-02) me
# - fixed a bug in function 'add_job' - an already defined
#   timer causes no deletion of old jobs
# - added function 'no_wakeup' - it deletes all timers for a
#   host (only if timestamp = 0)
# - added command line option 'nicelevel' and code for changing
#   the process scheduling priority
# - added warning, if nicelevel is less than 0
# - added pid for syslog messages
#
# Version 0.8   (2008-02-01) me
# - fixed a lot of typos :-/
# - added constants for return codes
# - added check for connected clients, so nmap cannot
#   kill my child again ;-)
# - changed pid file handling, should be more secure now
# - added security warning, if bound address is 0.0.0.0
# - removed code for controlling 'at' in some functions
# - splitted function for time calculations into 'wakeup1' and
#   'wakeup2'
# - changed time calculations a bit to be more exactly, because
#   'at' command can't handle seconds
# - removed function 'add_wakeup'
# - added function 'delete_job'
# - added function 'add_job'
# - added function 'read_jobs'
# - removed caching functions
# - added code to delete old wakeup timestamps (see below)
# - added code to change the wakeup timestamp (i.e. on timer
#   changes)
# - some code cleanups
#
# Version 0.7   (2008-01-30) me
# - moved check for uid 0 below commandline parsing function,
#   because '--help' should be possible as regular user :-|
#
# Version 0.6   (2008-01-23) me
# - support second client timestamp (optional)
# - added clock drift calculation
# - changed function clean_cache to work with new drift calculation
# - added a warning message for drifts greater than 24 hours
#
# Version 0.5   (2008-01-23) me
# - untaint for variables and path environment
# - added '-T' switch (Taint-Mode)
#
# Version 0.4   (2008-01-22) me
# - changed script name to 'wakemeup-server.pl' and pidfile
#   name to 'wakemeup-server.pid'
#
# Version 0.3   (2008-01-21) me
# - complete rewrite :-/
# - added some command line options and a parser for it
# - added logging via syslog or stdout/stderr
# - added daemon mode
# - added a cache
# - ToDo: - support second timestamp for a simple drift calculation
#         - better shutdown-hook script
#         - Debian Package
#
# Version 0.2   (2008-01-15) me
# - changed useless option 'LocalAddress' to 'LocalHost'
#
# Version 0.1   (2008-01-14) me
# - initial version
#
#####################################################################
